@@ -108,6 +108,7 @@ G_DEFINE_TYPE (GdmSimpleSlave, gdm_simple_slave, GDM_TYPE_SLAVE)
 static void create_new_session (GdmSimpleSlave *slave);
 static void destroy_session    (GdmSimpleSlave *slave);
 static void start_greeter      (GdmSimpleSlave *slave);
+static void start_session      (GdmSimpleSlave *slave);
 static void queue_start_session (GdmSimpleSlave *slave,
                                  const char     *service_name);
 
@@ -212,7 +213,9 @@ on_session_exited (GdmSession     *session,
                    GdmSimpleSlave *slave)
 {
         g_debug ("GdmSimpleSlave: session exited with code %d\n", exit_code);
-        gdm_slave_stopped (GDM_SLAVE (slave));
+        if (slave->priv->start_session_service_name == NULL) {
+                gdm_slave_stopped (GDM_SLAVE (slave));
+        }
 }
 
 static void
@@ -224,7 +227,9 @@ on_session_died (GdmSession     *session,
                  signal_number,
                  g_strsignal (signal_number));
 
-        gdm_slave_stopped (GDM_SLAVE (slave));
+        if (slave->priv->start_session_service_name == NULL) {
+                gdm_slave_stopped (GDM_SLAVE (slave));
+        }
 }
 
 static gboolean
@@ -443,17 +448,35 @@ stop_greeter (GdmSimpleSlave *slave)
         g_free (username);
 
         gdm_welcome_session_stop (GDM_WELCOME_SESSION (slave->priv->greeter));
-        gdm_greeter_server_stop (slave->priv->greeter_server);
+}
 
-        g_object_unref (slave->priv->greeter);
-        slave->priv->greeter = NULL;
+static void
+start_session (GdmSimpleSlave *slave)
+{
+        char    *auth_file;
+        auth_file = NULL;
+        add_user_authorization (slave, &auth_file);
+
+        g_assert (auth_file != NULL);
+
+        g_object_set (slave->priv->session,
+                      "user-x11-authority-file", auth_file,
+                      NULL);
+
+        g_free (auth_file);
+
+        gdm_session_start_session (GDM_SESSION (slave->priv->session),
+                                   slave->priv->start_session_service_name);
+
+        slave->priv->start_session_id = 0;
+        g_free (slave->priv->start_session_service_name);
+        slave->priv->start_session_service_name = NULL;
 }
 
 static gboolean
 start_session_timeout (GdmSimpleSlave *slave)
 {
 
-        char    *auth_file;
         gboolean migrated;
 
         g_debug ("GdmSimpleSlave: accredited");
@@ -469,28 +492,19 @@ start_session_timeout (GdmSimpleSlave *slave)
                    user switching. */
                 queue_greeter_reset (slave);
 
-                goto out;
+                slave->priv->start_session_id = 0;
+                g_free (slave->priv->start_session_service_name);
+                slave->priv->start_session_service_name = NULL;
+        } else {
+                if (slave->priv->greeter == NULL) {
+                        /* auto login */
+                        start_session (slave);
+                } else {
+                        /* Session actually gets started from on_greeter_session_stop */
+                        stop_greeter (slave);
+                }
         }
 
-        stop_greeter (slave);
-
-        auth_file = NULL;
-        add_user_authorization (slave, &auth_file);
-
-        g_assert (auth_file != NULL);
-
-        g_object_set (slave->priv->session,
-                      "user-x11-authority-file", auth_file,
-                      NULL);
-
-        g_free (auth_file);
-
-        gdm_session_start_session (GDM_SESSION (slave->priv->session),
-                                   slave->priv->start_session_service_name);
- out:
-        slave->priv->start_session_id = 0;
-        g_free (slave->priv->start_session_service_name);
-        slave->priv->start_session_service_name = NULL;
         return FALSE;
 }
 
@@ -736,6 +750,7 @@ create_new_session (GdmSimpleSlave *slave)
         char          *display_name;
         char          *display_hostname;
         char          *display_device;
+        char          *display_seat_id;
         char          *display_x11_authority_file;
 
         g_debug ("GdmSimpleSlave: Creating new session");
@@ -746,6 +761,7 @@ create_new_session (GdmSimpleSlave *slave)
                       "display-hostname", &display_hostname,
                       "display-is-local", &display_is_local,
                       "display-x11-authority-file", &display_x11_authority_file,
+                      "display-seat-id", &display_seat_id,
                       NULL);
 
         display_device = NULL;
@@ -757,6 +773,7 @@ create_new_session (GdmSimpleSlave *slave)
                                                        display_name,
                                                        display_hostname,
                                                        display_device,
+                                                       display_seat_id,
                                                        display_x11_authority_file,
                                                        display_is_local);
         g_free (display_id);
@@ -983,7 +1000,16 @@ on_greeter_session_stop (GdmGreeterSession *greeter,
                          GdmSimpleSlave    *slave)
 {
         g_debug ("GdmSimpleSlave: Greeter stopped");
-        gdm_slave_stopped (GDM_SLAVE (slave));
+        if (slave->priv->start_session_service_name == NULL) {
+                gdm_slave_stopped (GDM_SLAVE (slave));
+        } else {
+                gdm_greeter_server_stop (slave->priv->greeter_server);
+                start_session (slave);
+        }
+
+        g_object_unref (slave->priv->greeter);
+        slave->priv->greeter = NULL;
+
 }
 
 static void
@@ -992,7 +1018,9 @@ on_greeter_session_exited (GdmGreeterSession    *greeter,
                            GdmSimpleSlave       *slave)
 {
         g_debug ("GdmSimpleSlave: Greeter exited: %d", code);
-        gdm_slave_stopped (GDM_SLAVE (slave));
+        if (slave->priv->start_session_service_name == NULL) {
+                gdm_slave_stopped (GDM_SLAVE (slave));
+        }
 }
 
 static void
@@ -1009,10 +1037,11 @@ on_greeter_start_conversation (GdmGreeterServer *greeter_server,
                                const char       *service_name,
                                GdmSimpleSlave   *slave)
 {
-        g_debug ("GdmSimpleSlave: starting conversation with '%s' pam service'", service_name);
         if (slave->priv->greeter_reset_id > 0) {
                 return;
         }
+
+        g_debug ("GdmSimpleSlave: starting conversation with '%s' pam service'", service_name);
         gdm_session_start_conversation (GDM_SESSION (slave->priv->session),
                                         service_name);
 }
@@ -1183,6 +1212,9 @@ setup_server (GdmSimpleSlave *slave)
 
         /* Set the busy cursor */
         gdm_slave_set_busy_cursor (GDM_SLAVE (slave));
+
+        /* Set the initial keyboard layout to something reasonable */
+        gdm_slave_set_initial_keyboard_layout (GDM_SLAVE (slave));
 }
 
 static void
@@ -1405,11 +1437,13 @@ gdm_simple_slave_run (GdmSimpleSlave *slave)
 {
         char    *display_name;
         char    *auth_file;
+        char    *seat_id;
         gboolean display_is_local;
 
         g_object_get (slave,
                       "display-is-local", &display_is_local,
                       "display-name", &display_name,
+                      "display-seat-id", &seat_id,
                       "display-x11-authority-file", &auth_file,
                       NULL);
 
@@ -1419,7 +1453,7 @@ gdm_simple_slave_run (GdmSimpleSlave *slave)
                 gboolean res;
                 gboolean disable_tcp;
 
-                slave->priv->server = gdm_server_new (display_name, auth_file);
+                slave->priv->server = gdm_server_new (display_name, seat_id, auth_file);
 
                 disable_tcp = TRUE;
                 if (gdm_settings_client_get_boolean (GDM_KEY_DISALLOW_TCP,
