@@ -51,6 +51,11 @@
 #include <libxklavier/xklavier.h>
 #endif
 
+#ifdef WITH_SYSTEMD
+#include <systemd/sd-login.h>
+#include <systemd/sd-daemon.h>
+#endif
+
 #include "gdm-common.h"
 #include "gdm-xerrors.h"
 
@@ -541,8 +546,8 @@ gdm_slave_set_windowpath (GdmSlave *slave)
                 return;
         }
         if (XGetWindowProperty (slave->priv->server_display,
-                DefaultRootWindow (slave->priv->server_display), prop, 0, 1, 
-                False, AnyPropertyType, &actualtype, &actualformat, 
+                DefaultRootWindow (slave->priv->server_display), prop, 0, 1,
+                False, AnyPropertyType, &actualtype, &actualformat,
                 &nitems, &bytes_after, &buf)) {
                 g_debug ("no XFree86_VT property\n");
                 return;
@@ -1198,6 +1203,8 @@ _get_uid_and_gid_for_user (const char *username,
         return TRUE;
 }
 
+#ifdef WITH_CONSOLE_KIT
+
 static gboolean
 x11_session_is_on_seat (GdmSlave        *slave,
                         const char      *session_id,
@@ -1295,9 +1302,96 @@ x11_session_is_on_seat (GdmSlave        *slave,
         return ret;
 }
 
-char *
-gdm_slave_get_primary_session_id_for_user (GdmSlave   *slave,
-                                           const char *username)
+#endif
+
+#ifdef WITH_SYSTEMD
+static char*
+gdm_slave_get_primary_session_id_for_user_from_systemd (GdmSlave   *slave,
+                                                        const char *username)
+{
+        int     res, i;
+        char  **sessions;
+        uid_t   uid;
+        char   *primary_ssid;
+        gboolean got_primary_ssid;
+
+        primary_ssid = NULL;
+        got_primary_ssid = FALSE;
+
+        res = sd_seat_can_multi_session (slave->priv->display_seat_id);
+        if (res < 0) {
+                g_warning ("GdmSlave: Failed to determine whether seat is multi-session capable: %s", strerror (-res));
+                return NULL;
+        } else if (res == 0) {
+                g_debug ("GdmSlave: seat is unable to activate sessions");
+                return NULL;
+        }
+
+        if (! _get_uid_and_gid_for_user (username, &uid, NULL)) {
+                g_debug ("GdmSlave: unable to determine uid for user: %s", username);
+                return NULL;
+        }
+
+        res = sd_seat_get_sessions (slave->priv->display_seat_id, &sessions, NULL, NULL);
+        if (res < 0) {
+                g_warning ("GdmSlave: Failed to get sessions on seat: %s", strerror (-res));
+                return NULL;
+        }
+
+        if (sessions == NULL) {
+                g_debug ("GdmSlave: seat has no active sessions");
+                return NULL;
+        }
+
+        for (i = 0; sessions[i] != NULL; i++) {
+                char *type;
+                gboolean is_active;
+                gboolean is_x11;
+                uid_t other;
+
+                res = sd_session_get_type (sessions[i], &type);
+
+                if (res < 0) {
+                        g_warning ("GdmSlave: could not fetch type of session '%s': %s",
+                                   sessions[i], strerror (-res));
+                        continue;
+                }
+
+                is_x11 = g_strcmp0 (type, "x11") == 0;
+                g_free (type);
+
+                /* Only migrate to graphical sessions
+                 */
+                if (!is_x11) {
+                        continue;
+                }
+
+                /* Always give preference to non-active sessions,
+                 * so we migrate when we can and don't when we can't
+                 */
+                is_active = sd_session_is_active (sessions[i]) > 0;
+
+                res = sd_session_get_uid (sessions[i], &other);
+                if (res == 0 && other == uid && !got_primary_ssid) {
+                        g_free (primary_ssid);
+                        primary_ssid = g_strdup (sessions[i]);
+
+                        if (!is_active) {
+                                got_primary_ssid = TRUE;
+                        }
+                }
+                free (sessions[i]);
+        }
+
+        free (sessions);
+        return primary_ssid;
+}
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+static char *
+gdm_slave_get_primary_session_id_for_user_from_ck (GdmSlave   *slave,
+                                                   const char *username)
 {
         gboolean    res;
         gboolean    can_activate_sessions;
@@ -1309,11 +1403,6 @@ gdm_slave_get_primary_session_id_for_user (GdmSlave   *slave,
         int         i;
         uid_t       uid;
 
-        if (slave->priv->display_seat_id == NULL || slave->priv->display_seat_id[0] == '\0') {
-                g_debug ("GdmSlave: display seat ID is not set; can't switch sessions");
-                return NULL;
-        }
-
         manager_proxy = NULL;
         primary_ssid = NULL;
         sessions = NULL;
@@ -1396,11 +1485,100 @@ gdm_slave_get_primary_session_id_for_user (GdmSlave   *slave,
 
         return primary_ssid;
 }
+#endif
 
+char *
+gdm_slave_get_primary_session_id_for_user (GdmSlave   *slave,
+                                           const char *username)
+{
+
+        if (slave->priv->display_seat_id == NULL || slave->priv->display_seat_id[0] == '\0') {
+                g_debug ("GdmSlave: display seat ID is not set; can't switch sessions");
+                return NULL;
+        }
+
+#ifdef WITH_SYSTEMD
+        if (sd_booted () > 0) {
+                return gdm_slave_get_primary_session_id_for_user_from_systemd (slave, username);
+        }
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+        return gdm_slave_get_primary_session_id_for_user_from_ck (slave, username);
+#else
+        return NULL;
+#endif
+}
+
+#ifdef WITH_SYSTEMD
 static gboolean
-activate_session_id (GdmSlave   *slave,
-                     const char *seat_id,
-                     const char *session_id)
+activate_session_id_for_systemd (GdmSlave   *slave,
+                                 const char *seat_id,
+                                 const char *session_id)
+{
+        DBusError    local_error;
+        DBusMessage *message;
+        DBusMessage *reply;
+        gboolean     ret;
+
+        ret = FALSE;
+        reply = NULL;
+
+        /* Can't activate what's already active. We want this
+         * to fail, because we don't want migration to succeed
+         * if the only active session is the one just created
+         * at the login screen.
+         */
+        if (sd_session_is_active (session_id) > 0) {
+                return FALSE;
+        }
+
+        dbus_error_init (&local_error);
+
+        message = dbus_message_new_method_call ("org.freedesktop.login1",
+                                                "/org/freedesktop/login1",
+                                                "org.freedesktop.login1.Manager",
+                                                "ActivateSessionOnSeat");
+        if (message == NULL) {
+                goto out;
+        }
+
+        if (! dbus_message_append_args (message,
+                                        DBUS_TYPE_STRING, &session_id,
+                                        DBUS_TYPE_STRING, &seat_id,
+                                        DBUS_TYPE_INVALID)) {
+                goto out;
+        }
+
+        reply = dbus_connection_send_with_reply_and_block (dbus_g_connection_get_connection (slave->priv->connection),
+                                                           message,
+                                                           -1,
+                                                           &local_error);
+        if (dbus_error_is_set (&local_error)) {
+                g_warning ("GdmSlave: Unable to activate session: %s", local_error.message);
+                dbus_error_free (&local_error);
+                goto out;
+        }
+
+        ret = TRUE;
+
+ out:
+        if (message != NULL) {
+                dbus_message_unref (message);
+        }
+        if (reply != NULL) {
+                dbus_message_unref (reply);
+        }
+
+        return ret;
+}
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+static gboolean
+activate_session_id_for_ck (GdmSlave   *slave,
+                            const char *seat_id,
+                            const char *session_id)
 {
         DBusError    local_error;
         DBusMessage *message;
@@ -1448,16 +1626,82 @@ activate_session_id (GdmSlave   *slave,
 
         return ret;
 }
+#endif
 
 static gboolean
-session_unlock (GdmSlave   *slave,
-                const char *ssid)
+activate_session_id (GdmSlave   *slave,
+                     const char *seat_id,
+                     const char *session_id)
+{
+
+#ifdef WITH_SYSTEMD
+        if (sd_booted () > 0) {
+                return activate_session_id_for_systemd (slave, seat_id, session_id);
+        }
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+        return activate_session_id_for_ck (slave, seat_id, session_id);
+#else
+        return FALSE;
+#endif
+}
+
+#ifdef WITH_SYSTEMD
+static gboolean
+session_unlock_for_systemd (GdmSlave   *slave,
+                            const char *ssid)
+{
+        DBusError       error;
+        DBusMessage    *message;
+        DBusMessage    *reply;
+
+        dbus_error_init (&error);
+
+        message = dbus_message_new_method_call ("org.freedesktop.login1",
+                                                "/org/freedesktop/login1",
+                                                "org.freedesktop.login1.Manager",
+                                                "UnlockSession");
+
+        if (message == NULL) {
+                g_debug ("GdmSlave: couldn't allocate the D-Bus message");
+                return FALSE;
+        }
+
+        if (! dbus_message_append_args (message,
+                                        DBUS_TYPE_STRING, &ssid,
+                                        DBUS_TYPE_INVALID)) {
+                g_debug ("GdmSlave: couldn't attach the D-Bus message data");
+                return FALSE;
+        }
+
+        reply = dbus_connection_send_with_reply_and_block (dbus_g_connection_get_connection (slave->priv->connection),
+                                                           message,
+                                                           -1, &error);
+        dbus_message_unref(message);
+        if (reply != NULL) {
+                dbus_message_unref(reply);
+        }
+
+        if (dbus_error_is_set (&error)) {
+                g_warning ("GdmSlave: Unable to unlock session: %s", error.message);
+                dbus_error_free (&error);
+                return FALSE;
+        }
+
+        return TRUE;
+}
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+static gboolean
+session_unlock_for_ck (GdmSlave   *slave,
+                       const char *ssid)
 {
         DBusError       error;
         DBusMessage    *message;
         DBusMessage    *reply;
 
-        g_debug ("ConsoleKit: Unlocking session %s", ssid);
         message = dbus_message_new_method_call (CK_NAME,
                                                 ssid,
                                                 CK_SESSION_INTERFACE,
@@ -1485,6 +1729,27 @@ session_unlock (GdmSlave   *slave,
 
         return TRUE;
 }
+#endif
+
+static gboolean
+session_unlock (GdmSlave   *slave,
+                const char *ssid)
+{
+
+        g_debug ("Unlocking session %s", ssid);
+
+#ifdef WITH_SYSTEMD
+        if (sd_booted () > 0) {
+                return session_unlock_for_systemd (slave, ssid);
+        }
+#endif
+
+#ifdef WITH_CONSOLE_KIT
+        return session_unlock_for_ck (slave, ssid);
+#else
+        return TRUE;
+#endif
+}
 
 gboolean
 gdm_slave_switch_to_user_session (GdmSlave   *slave,
